package kepler;

import kepler.view.*;
import kepler.object.*;
import kepler.control.*;

import utils.d2.*;
import utils.timers.*;
import utils.GUI.GridBagHelpers;
import utils.GUI.Bar;

import webwisdom.tango.*;
import netscape.javascript.JSObject;

import java.awt.*;
import java.awt.Dimension;
import java.applet.Applet;
import java.net.URL;
import java.net.MalformedURLException;
import java.io.*;

/***********************************************************************
 *  a tangoized version of the kepler applet.
 *  relevant areas can be located by searching for TANGO
 *  this class, and EventHandles should be all that are necessary to
 *  understand for the port.
 ***********************************************************************/
public class Kepler extends Applet implements TDataListener, TControlListener, kepler.EventHandles {

	// TANGO variables
	TAgent agent = null;
	boolean	isMaster = true;

	private int channel = 50;
	private boolean javaOnly = true;
	boolean inTango = true;
	boolean fromScript = false;

	//TANGO protocol
	boolean init_completed = false;
	boolean master_init_handled = false;
	boolean incoming_event = false;

	KeplerView view;	// large black area where things are drawn

	ConfigPanel config;	// these three are the control panels
	GeneralPanel general;	// on right of applet
	SweepPanel sweep;

	TimeAreaView taview;	 // display of time area pairs

	Planet earth;
	Star sun;
	TwoPerspective p;	// used in drawing the planets

	FixedKicker time;	// provides simulation time
	AnimationThread film,ofilm; // force periodic update of view/taview
	boolean running;

	SSystemConfig parameters; // hold on to the parameters defining the orbit

	public final byte TP_START_SWEEPING=0;
	public final byte TP_STOP_SWEEPING=1;
	public final byte TP_PAUSE=2;
	public final byte TP_CONTINUE=3;
	public final byte TP_ERASE_AREAS=4;
	public final byte TP_ERASE_LINES=5;
	public final byte TP_ZOOM_IN=6;
	public final byte TP_ZOOM_OUT=7;
	public final byte TP_SHOW_VEL=8;
	public final byte TP_NO_SHOW_VEL=9;
	public final byte TP_SHOW_ACC=10;
	public final byte TP_NO_SHOW_ACC=11;
	public final byte TP_SHOW_ORB=12;
	public final byte TP_NO_SHOW_ORB=13;
	public final byte TP_SWITCH_CONTROLS=14;
	public final byte TP_MOUSE_DRAG=15;
	public final byte TP_MOUSE_DOWN=16; 
	public final byte TP_SET_MASS=17;
	public final byte TP_SET_SM=18;
	public final byte TP_SET_ECC=19;

	public void init() {
		initThis();
		initTango();
		initApplet();
	}

	public void initTango() {

		String arg = null;
		try {
			arg = getParameter("interface");
			inTango = true;
		}
		catch(Exception e) {
			inTango = false;
		}

		if (arg == null) {
			inTango = false;
		}
		System.out.println("<Kepler> arg : " + arg);
		System.out.println("<Kepler> inTango : " + inTango);

		if (inTango) {
			setTango(arg);
		}
	}

	public void initThis() {
		FixedKicker.dt=0.005;

		int viewW=350;
		int viewH=300;
		setBackground(new Color(60,60,125));
		setForeground(Color.white);

		URL image_base=constructImageBase();
		System.out.println("image base : " + image_base);
		sun=new Star(new TwoVector(),1.0,image_base);
		earth=new Planet(sun,1.0,0.0,image_base);
		
		parameters=new SSystemConfig();
		parameters.stellarmass=1.0;
		parameters.semimajor=1.0;
		parameters.eccentricity=0.0;

		p=new TwoPerspective( new Dimension(viewW,viewH), 4.0);
		view=new KeplerView(sun,earth,p);
		view.resize(viewW,viewH);

		config=new ConfigPanel();
		general=new GeneralPanel();
		sweep=new SweepPanel();

		taview=new TimeAreaView(earth);
		myLayout();

		running=false;

	}

	public void initApplet() {

		init_completed = true;
		if (!master_init_handled) {
			if (isMaster) {
				enableAWT(); 
			}
			else {
				disableAWT();
			}
			master_init_handled = true;
		}

		if (isMaster) {
			if (init_completed) {
				enableAWT();
				master_init_handled = true;
			}
		}
		else {
			if (init_completed) {
				disableAWT();
				master_init_handled = true;
			}
		}
	}

	public void startThis() {
		System.out.println("<Kepler> startThis()");
	}

	public void commJava() {

		try {
			agent = new TAgent(new TLAgentApplet(this));
		}
		catch(TangoException e) {
			e.printStackTrace();
			System.out.println("<Kepler> agent is not proprely assigned");
		}

	}

	public void commJavaScript() {

		JSObject jsWin = JSObject.getWindow(this);
		Object[] args = new Object[1];
		args[0] = this;
		jsWin.call("appletReadyJS", args);
	}

	public void setTango(String arg) {
                if ( arg.equals("javaApplet") ) {
			fromScript = false;
			commJava();
		}
		else if ( arg.equals("javaScript") ) {
			fromScript = true;
			commJavaScript();
		}

		while ( agent == null ) {
			System.out.println("<Kepler> waiting for the agent to be assigned");
		}
		System.out.println("<Kepler> agent assigned");

		//dataListener = agent.addTDataListener(channel, this);
		//controlListener = agent.addTControlListener(this);
		agent.addTDataListener(channel, this);
		agent.addTControlListener(this);
		isMaster = agent.isMaster();
	}

	public void setTango(String args[]) {

		try {
			agent = new TAgent(new TLAgentApplication(args));
		}
		catch(TangoException e) {
			e.printStackTrace();
			System.out.println("<Kepler> agent is not proprely assigned");
		}

		System.out.println("<Kepler> Application2 agent assigned");

		agent.addTDataListener(channel, this);
		System.out.println("<Kepler> Data added");

		agent.addTControlListener(this);
		System.out.println("<Kepler> Control added");

//		isMaster = agent.isMaster();
		System.out.println("<Kepler> Data Control added");
	}

	public void initTangoJS(TAgent a, Number ch, Number ma) {

		channel = ((Number) ch).intValue();

		if ( ((Number) ma).intValue() ==  0 ) {
			isMaster = true;
		}
		else {
			isMaster = false;
		}

		agent = a;
	}

	public void masterChangedJS(Number ma, String masterName) {
		if ( ((Number) ma).intValue() ==  0 ) {
			isMaster = true;
		}
		else {
			isMaster = false;
		}

		if (isMaster) {
			enableAWT(); 
		}
		else {
			disableAWT();
		}
	}

	public void masterChanged(boolean master, String masterName) {
		isMaster = master;

		if (isMaster) {
			enableAWT(); 
		}
		else {
			disableAWT();
		}
	}

	public void participantJoined(String participantName) {
	}

	public void participantLeft(String participantName) {
	}

	public void audioRequested(boolean isREquested) {
	}

	private byte[] stringToBytes(String s) {
		int l = s.length();
		byte[] t = new byte[l];
		s.getBytes(0, l, t, 0);
		return t;
	}

	private String bytesToString(byte t[]) {
		return new String(t, 0);
	}

	public void destroy() {
		System.out.println("<Kepler> inside destroy()");

		if ( fromScript ) {
			agent.removeTControlListener(this);
			agent.removeTDataListener(this);
			System.out.println("<Kepler> data and control removed");
			return;
		}

		if (agent != null) {
			agent.exit();
		}
		super.destroy();
	}

	void myLayout(){

		setLayout(new GridBagLayout());
		GridBagHelpers.insert(
				this,view,
				1,1,
				1,4,
				GridBagConstraints.BOTH,
				GridBagConstraints.NORTHWEST,
				0,0,
				1.0,1.0,
				8,8,0,0				
				);
		GridBagHelpers.insert(
				this,general,
				2,1,
				1,1,
				GridBagConstraints.VERTICAL,
				GridBagConstraints.CENTER,
				0,0,
				0.0,1.0,
				8,0,5,8
				);
		GridBagHelpers.insert_fill(
				this,new Bar(),
				2,2,
				1,1,
				5,0,5,8
				);
		GridBagHelpers.insert(
				this,config,
				2,3,
				1,1,
				GridBagConstraints.VERTICAL,
				GridBagConstraints.CENTER,
				0,0,
				0.0,2.0,
				5,5,5,8
				);
		GridBagHelpers.insert(
				this,new Bar(),
				2,4,	
				1,1,
				GridBagConstraints.HORIZONTAL,
				GridBagConstraints.SOUTH,
				0,0,
				0.0,0.0,
				5,0,5,8
				);

		GridBagHelpers.insert(
				this,taview,
				1,5,
				1,1,
				10,8,8,5,
				GridBagConstraints.CENTER
				);
		GridBagHelpers.insert(
				this,sweep,
				2,5,
				1,1,
				10,5,8,8,
				GridBagConstraints.CENTER
				);
	}

	void startThreads() {
		if (!running) {
			film=new AnimationThread(view,75);
			ofilm=new AnimationThread(taview,75);
			time=new FixedKicker(earth,75);
			time.start();
			film.start();
			ofilm.start();
			running=true;
		}
	}

	void stopThreads() {
		if (running) {
			film.pleaseStop();
			ofilm.pleaseStop();
			time.pleaseStop();
			running=false;
		}
	}

	public boolean action(Event e,Object arg) {
		if ( arg instanceof String) {
			if ( isMaster ) {
				tangoizeEvent(e,0,0);
			}
			return stringAction(e,(String)arg);
		}
		else if (arg instanceof SSystemConfig) {
			stopThreads();
			handleEvent(new Event(this,Event.ACTION_EVENT,STOP_SWEEPING));
			distributeEvent(new Event(this,Event.ACTION_EVENT,ERASE_AREAS));
			SSystemConfig ss=(SSystemConfig)arg;
			sun.setMass(ss.stellarmass);
			earth.defineOrbit(sun,ss.semimajor,ss.eccentricity);
			if ( isMaster ) {
				sendNewMass(ss.stellarmass);
				sendNewSemiMajor(ss.semimajor);
				sendNewEcc(ss.eccentricity);
			}
			return distributeEvent(e);
		}
		else return super.action(e,arg);
	}

	boolean stringAction(Event e,String sarg) {
		// can get signals which affect threads
		if(PAUSE.equals(sarg) ) {
			stopThreads();	return true;
		}
		else if (CONTINUE.equals(sarg) ) {
			startThreads(); return true;
		}
		// if sarg starts with "Zoom" -- then it affects the 
		// view's perspective so send it there
		else if (sarg.startsWith("Zoom") ||
			 sarg.equals(SHOW_VEL)	 ||
			 sarg.equals(NO_SHOW_VEL)||
			 sarg.equals(SHOW_ACC)	 ||
			 sarg.equals(SHOW_ORB)	 ||
			 sarg.equals(NO_SHOW_ORB)||
			 sarg.equals(NO_SHOW_ACC) ) {
			 return view.handleEvent(e);
		}
		else if ( START_SWEEPING.equals(sarg)) {
			earth.startSweeping();
			view.readEvent(e);
			sweep.readEvent(e);
			taview.readEvent(e);
			return true;
		}
		else if ( STOP_SWEEPING.equals(sarg)) {
			earth.stopSweeping();
			view.readEvent(e);
			sweep.readEvent(e);
			taview.readEvent(e);
			return true;
		}
		else if (ERASE_AREAS.equals(sarg)) {
			System.out.println("<Kepler> EA 1");
			earth.stopSweeping();
			System.out.println("<Kepler> EA 2");
			boolean tmp=distributeEvent(e);
			System.out.println("<Kepler> EA 3");
			if(!film.isAlive()) {
				System.out.println("<Kepler> EA 4");
				view.draw();
				System.out.println("<Kepler> EA 5");
			}
			System.out.println("<Kepler> EA 6");
			return tmp;
		}
		else if( ERASE_LINES.equals(sarg)) {
			boolean tmp=distributeEvent(e);
			if( !running ) {
				view.draw();
			}
			return tmp;
		}
		else return false;
	}

	public boolean mouseDown(Event e,int x,int y) {
		if ( e.target==view && !running ) {
			view.draw();
			if (!(e.arg instanceof SSystemConfig || incoming_event)) {
				tangoizeEvent(e,x,y);
			}
			return true;
		}
		else return super.mouseDown(e,x,y);
	}
	
	public boolean mouseDrag(Event e,int x,int y) {
		if ( e.target==view && !running) {
			view.draw();
			if (!(e.arg instanceof SSystemConfig || incoming_event)) {
				tangoizeEvent(e,x,y);
			}
			return true;
		}
		else return super.mouseDrag(e,x,y);
	}

	boolean distributeEvent(Event e) {
		boolean tmp=true;
		tmp=view.readEvent(e)|taview.readEvent(e)|
			general.readEvent(e)|sweep.readEvent(e)|config.readEvent(e);
		return tmp;
	}

	public boolean readEvent(Event e) {
		return handleEvent(e)||distributeEvent(e);
	}
	
	public void stop(){
		stopThreads();
	}

	URL constructImageBase() {
		URL tmp=null;
		String arg=getParameter("image_offset");
		try {
			if( arg != null) {
				tmp=new URL(getCodeBase(),arg);
			}
			else {
				tmp=new URL(getCodeBase(),"kepler/images/");
			}
			System.out.println("code base : " + getCodeBase());
		}
		catch(MalformedURLException mue) {
			System.out.println("unable to locate images...");
			System.out.println("code_base "+getCodeBase());
			System.out.println("supplied offset "+arg);
			System.out.println("default offset "+"kepler/images/");
			return null;
		}
		return tmp;
	}

	/***************************************************
	 *	TANGO 
	 *	first look at recieving messages
	 ***************************************************/
	public void receive(byte m[]) {
		receiveData(m);
	}

/**
	void receiveControl(AppEventMessage m) {
		CAppAnswerEventMessage mes =	(CAppAnswerEventMessage) m;

		switch((int) mes.getAnswerType()) {
		case CAConsts.ISMASTER:
			isMaster = mes.isMaster();
			if (isMaster) {
				if (init_completed) {
					enableAWT();
					master_init_handled = true;
				}
			}
			else {
				if (init_completed) {
					disableAWT();
					master_init_handled = true;
				}
			}
			break;
		}
	}
**/

	void receiveData(byte m[]) {
		ByteArrayInputStream bis = new ByteArrayInputStream(m);
		switch((int) bis.read()) {
		case TP_START_SWEEPING:
			stringAction(new Event( this, Event.ACTION_EVENT, START_SWEEPING), START_SWEEPING);
			break;
		case TP_STOP_SWEEPING:
			stringAction(new Event( this, Event.ACTION_EVENT, STOP_SWEEPING), STOP_SWEEPING); 
			break;
		case TP_PAUSE:
			stringAction(new Event( this, Event.ACTION_EVENT, PAUSE), PAUSE); 
			break;
		case TP_CONTINUE:
			stringAction(new Event( this, Event.ACTION_EVENT, CONTINUE), CONTINUE); 
			break;
		case TP_ERASE_AREAS:
			stringAction(new Event( this, Event.ACTION_EVENT, ERASE_AREAS), ERASE_AREAS);	
			break;
		case TP_ERASE_LINES:
			stringAction(new Event( this, Event.ACTION_EVENT, ERASE_LINES), ERASE_LINES); 
			break;
		case TP_ZOOM_IN:
			stringAction(new Event( this, Event.ACTION_EVENT, ZOOM_IN), ZOOM_IN); 
			break;
		case TP_ZOOM_OUT:
			stringAction(new Event( this, Event.ACTION_EVENT, ZOOM_OUT), ZOOM_OUT); 
			break;
		case TP_SHOW_VEL:
			stringAction(new Event( this, Event.ACTION_EVENT, SHOW_VEL),SHOW_VEL ); 
			break;
		case TP_NO_SHOW_VEL:
			stringAction(new Event( this, Event.ACTION_EVENT, NO_SHOW_VEL), NO_SHOW_VEL); 
			break;
		case TP_SHOW_ACC:
			stringAction(new Event( this, Event.ACTION_EVENT, SHOW_ACC), SHOW_ACC); 
			break;
		case TP_NO_SHOW_ACC:
			stringAction(new Event( this, Event.ACTION_EVENT, NO_SHOW_ACC), NO_SHOW_ACC); 
			break;
		case TP_SHOW_ORB:
			stringAction(new Event( this, Event.ACTION_EVENT, SHOW_ORB), SHOW_ORB); 
			break;
		case TP_NO_SHOW_ORB:
			stringAction(new Event( this, Event.ACTION_EVENT, NO_SHOW_ORB), NO_SHOW_ORB); 
			break;
		case TP_SWITCH_CONTROLS:
			stringAction(new Event( this, Event.ACTION_EVENT, SWITCH_CONTROLS), SWITCH_CONTROLS); 
			break;
		case TP_MOUSE_DRAG: 
			DataInputStream dis = new DataInputStream(bis);
			try {
				int x = dis.readInt();
				int y = dis.readInt();
				incoming_event = true;
				distributeEvent(new Event( view, 0, Event.MOUSE_DRAG, x, y, 0, 0)); 
				incoming_event = false;
			}
			catch (Exception ex) {}
			break;
		case TP_MOUSE_DOWN: 
			dis = new DataInputStream(bis);
			try {
				int x = dis.readInt();
				int y = dis.readInt();
				incoming_event = true;
				distributeEvent(new Event( view, 0, Event.MOUSE_DOWN, x, y, 0, 0));
				incoming_event = false;
			}
			catch (Exception ex) {}
			break; 
		case TP_SET_MASS:
			dis = new DataInputStream(bis);
			try {
				double d = dis.readDouble();
				receiveNewMass(d); 
			}
			catch (Exception ex) {}
			break;
		case TP_SET_SM:
			dis = new DataInputStream(bis);
			try {
				double d = dis.readDouble();
				recieveNewSemiMajor(d); 
			}
			catch (Exception ex) {} 
			break;
		case TP_SET_ECC:
			dis = new DataInputStream(bis);
			try {
				double d = dis.readDouble();
				recieveNewEcc(d); 
			}
			catch (Exception ex) {}	
			break;
		}
	}

	/****************************************
	 * TANGO -- here do the sending part
	 ****************************************/

	void tangoizeEvent(Event e, int x, int y) {
		ByteArrayOutputStream bos = new ByteArrayOutputStream();
 
		if (e.id == Event.ACTION_EVENT) {
			if ( e.arg.equals(SHOW_ACC)) {
				bos.write(TP_SHOW_ACC);
			}
			else if ( e.arg.equals(START_SWEEPING)) {
				bos.write(TP_START_SWEEPING);
			}
			else if ( e.arg.equals(STOP_SWEEPING)) {
				bos.write(TP_STOP_SWEEPING);
			}
			else if ( e.arg.equals(PAUSE)) {
				bos.write(TP_PAUSE);
			}
			else if ( e.arg.equals(CONTINUE)) {
				bos.write(TP_CONTINUE);
			}
			else if ( e.arg.equals(ERASE_AREAS)) {
				bos.write(TP_ERASE_AREAS);
			}
			else if ( e.arg.equals(ERASE_LINES)) {
				bos.write(TP_ERASE_LINES);
			}
			else if ( e.arg.equals(ZOOM_IN)) {
				bos.write(TP_ZOOM_IN);
			}
			else if ( e.arg.equals(ZOOM_OUT)) {
				bos.write(TP_ZOOM_OUT);
			}
			else if ( e.arg.equals(SHOW_VEL)) {
				bos.write(TP_SHOW_VEL);
			}
			else if ( e.arg.equals(NO_SHOW_VEL)) {
				bos.write(TP_NO_SHOW_VEL);
			}
			else if ( e.arg.equals(SHOW_ORB)) {
				bos.write(TP_SHOW_ORB);
			}
			else if ( e.arg.equals(NO_SHOW_ORB)) {
				bos.write(TP_NO_SHOW_ORB);
			}
			else if ( e.arg.equals(SWITCH_CONTROLS)) {
				bos.write(TP_SWITCH_CONTROLS);
			}
			else if ( e.arg.equals(NO_SHOW_ACC)) {
				bos.write(TP_NO_SHOW_ACC);
			}
		}
		else if (e.id == Event.MOUSE_DOWN) {
			bos.write(TP_MOUSE_DOWN);
			DataOutputStream dos = new DataOutputStream(bos);
			try {
				dos.writeInt(e.x);
				dos.writeInt(e.y);
			} 
			catch (Exception ex) {}
		}
		else if (e.id == Event.MOUSE_DRAG) {
			bos.write(TP_MOUSE_DRAG);
			DataOutputStream dos = new DataOutputStream(bos);
			try {
				dos.writeInt(e.x);
				dos.writeInt(e.y);
			} 
			catch (Exception ex) {}
		}

		byte[] buffer = bos.toByteArray();
		sendMessage(buffer);
	}

	/*************************************
	 * TANGO -- the functions which you have requested

	 * 
	 *************************************/

	public void sendNewMass(double new_mass) {
		ByteArrayOutputStream bos = new ByteArrayOutputStream();
		bos.write(TP_SET_MASS);
		DataOutputStream dos = new DataOutputStream(bos);
		try {
			dos.writeDouble(new_mass);
		} 
		catch (Exception ex) {}
		byte[] buffer = bos.toByteArray();
		sendMessage(buffer);
	}

	public void sendNewSemiMajor(double semi) {
		ByteArrayOutputStream bos = new ByteArrayOutputStream();
		bos.write(TP_SET_SM);
		DataOutputStream dos = new DataOutputStream(bos);
		try {
			dos.writeDouble(semi);
		} 
		catch (Exception ex) {}
		byte[] buffer = bos.toByteArray();
		sendMessage(buffer); 
	}

	public void sendNewEcc(double ecc) {
		ByteArrayOutputStream bos = new ByteArrayOutputStream();
		bos.write(TP_SET_ECC);
		DataOutputStream dos = new DataOutputStream(bos);
		try {
			dos.writeDouble(ecc);
		} 
		catch (Exception ex) {}
		byte[] buffer = bos.toByteArray();
		sendMessage(buffer);
	}

	public void sendMessage(byte sendData[]) {
		if (!inTango) {
			return;
		}

		agent.send(channel, sendData);
	}

	public void receiveNewMass(double m) {
		parameters.stellarmass=m;
		setNewConfiguration(parameters);
	}

	public void recieveNewSemiMajor(double semi) {
		parameters.semimajor=semi;
		setNewConfiguration(parameters);
	}
	public void recieveNewEcc(double ecc){
		parameters.eccentricity=ecc;
		setNewConfiguration(parameters);
	}

	void setNewConfiguration(SSystemConfig sc) {
		stopThreads();
		handleEvent(new Event(this,Event.ACTION_EVENT,STOP_SWEEPING));
		distributeEvent(new Event(this,Event.ACTION_EVENT,ERASE_AREAS));
		distributeEvent( new Event( this,Event.ACTION_EVENT,sc) );
		sun.setMass(sc.stellarmass);
		earth.defineOrbit(sun,sc.semimajor,sc.eccentricity);
	}

	void disableAWT() {
		config.disable();
		general.disable();
		sweep.disable();
		//view.disable();
	}
	
	void enableAWT() {
		config.enable();
		general.enable();
		sweep.enable();
		//view.enable();
	}

	public static void main(String args[]) {
		Frame f = new Frame("Kepler Applet");
		Kepler kEpler = new Kepler();

		kEpler.initThis();

		try {
			System.out.println("<Kepler> arg[0] : " + args[0]);
			kEpler.inTango = true;
		}
		catch (Exception e) {
			kEpler.inTango = false;
		}

		System.out.println("<Kepler> inTango : " + kEpler.inTango);

		if (kEpler.inTango) {
			kEpler.setTango(args);
		}

		kEpler.initApplet();

		f.add("Center", kEpler);
		f.setSize(550, 550);
		f.show();
	}
}
